/**************************************************************************
General Description:
This module implements the C-part of the USB driver. See usbdrv.h for a
documentation of the entire driver.
**************************************************************************/

#include "stdafx.h"
#include "usbconfig.h"
#include "usbdrv.h"
#include "usb_descriptors.h"

/* raw USB registers / interface to assembler code: */
/* usbRxBuf MUST be in 1 byte addressable range (because usbInputBuf is only 1 byte) */
__no_init UCHAR usbRxBuf[2][USB_BUFSIZE] __attribute__ ((section (USB_BUFFER_SECTION))) IAR_SECTION(USB_BUFFER_SECTION);/* raw RX buffer: PID, 8 bytes data, 2 bytes CRC */
UCHAR       usbDeviceAddr;      /* assigned during enumeration, defaults to 0 */
UCHAR       usbNewDeviceAddr;   /* device ID which should be set after status phase */
UCHAR       usbConfiguration;   /* currently selected configuration. Administered by driver, but not used */
UCHAR       usbInputBuf;        /* ptr to raw buffer used for receiving */
UCHAR       usbAppBuf;          /* ptr to raw buffer passed to app for processing */
volatile SCHAR usbRxLen;        /* = 0; number of bytes in usbAppBuf; 0 means free */
UCHAR       usbCurrentTok;      /* last token received, if more than 1 rx endpoint: MSb=endpoint */
UCHAR       usbRxToken;         /* token for data we received; if more than 1 rx endpoint: MSb=endpoint */
UCHAR       usbMsgLen = 0xff;   /* remaining number of bytes, no msg to send if -1 (see usbMsgPtr) */
volatile SCHAR usbTxLen = -1;   /* number of bytes to transmit with next IN token */
UCHAR       usbTxBuf[USB_BUFSIZE];/* data to transmit with next IN, free if usbTxLen == -1 */
#if USB_CFG_HAVE_INTRIN_ENDPOINT
volatile SCHAR usbTxLen1 = -1;  /* TX count for endpoint 1 */
UCHAR       usbTxBuf1[USB_BUFSIZE];/* TX data for endpoint 1 */
#if USB_CFG_HAVE_INTRIN_ENDPOINT3
volatile SCHAR usbTxLen3 = -1;  /* TX count for endpoint 1 */
UCHAR       usbTxBuf3[USB_BUFSIZE];/* TX data for endpoint 1 */
#endif
#endif
UCHAR       usbAckBuf[1] = {USBPID_ACK};    /* transmit buffer for ack tokens */
UCHAR       usbNakBuf[1] = {USBPID_NAK};    /* transmit buffer for nak tokens */

/* USB status registers / not shared with asm code */
UCHAR           *usbMsgPtr;     /* data to transmit next -- ROM or RAM address */
static UCHAR    usbMsgFlags;    /* flag values see below */
static UCHAR    usbIsReset;     /* = 0; USB bus is in reset phase */

#define USB_FLG_TX_PACKET       (1<<0)
/* Leave free 6 bits after TX_PACKET. This way we can increment usbMsgFlags to toggle TX_PACKET */
#define USB_FLG_MSGPTR_IS_ROM   (1<<6)
#define USB_FLG_USE_DEFAULT_RW  (1<<7)

#ifndef PRG_RDB
#define PRG_RDB(addr)   pgm_read_byte(addr)
#endif

/**************************************************************************
We use this union to do type conversions. This is better optimized than
type casts in gcc 3.4.3 and much better than using bit shifts to build
ints from chars. Byte ordering is not a problem on an 8 bit platform.
**************************************************************************/
typedef union
{
    UINT	word;
    UCHAR   *ptr;
    UCHAR   bytes[2];

} converter_t;

#if USB_CFG_HAVE_INTRIN_ENDPOINT
static UCHAR    usbTxPacketCnt1;
#if USB_CFG_IMPLEMENT_HALT
static UCHAR    usbHalted1;         /* not 0 if endpoint 1 is halted */
#endif

/**************************************************************************

**************************************************************************/
void usbSetInterrupt(UCHAR *data, UCHAR len)
{
	UCHAR *p, i;

#if USB_CFG_IMPLEMENT_HALT
    if(usbHalted1)
        return;
#endif

    i = USBPID_DATA1;
    if(usbTxPacketCnt1 & 1)
        i = USBPID_DATA0;
    if(usbTxLen1 < 0){      /* packet buffer was empty */
        usbTxPacketCnt1++;
    }else{
        usbTxLen1 = -1;     /* avoid sending incomplete interrupt data */
    }
    p = usbTxBuf1;
    *p++ = i;
    for(i=len;i--;)
        *p++ = *data++;
    usbCrc16Append(&usbTxBuf1[1], len);
    usbTxLen1 = len + 4;    /* len must be given including sync byte */
}
#endif

#if USB_CFG_HAVE_INTRIN_ENDPOINT3
static UCHAR    usbTxPacketCnt3;

void    usbSetInterrupt3(UCHAR *data, UCHAR len)
{
UCHAR       *p, i;

    i = USBPID_DATA1;
    if(usbTxPacketCnt3 & 1)
        i = USBPID_DATA0;
    if(usbTxLen3 < 0){      /* packet buffer was empty */
        usbTxPacketCnt3++;
    }else{
        usbTxLen3 = -1;     /* avoid sending incomplete interrupt data */
    }
    p = usbTxBuf3;
    *p++ = i;
    for(i=len;i--;)
        *p++ = *data++;
    usbCrc16Append(&usbTxBuf3[1], len);
    usbTxLen3 = len + 4;    /* len must be given including sync byte */
}
#endif

static UCHAR    usbRead(UCHAR *data, UCHAR len)
{
#if USB_CFG_IMPLEMENT_FN_READ
    if(usbMsgFlags & USB_FLG_USE_DEFAULT_RW){
#endif
        UCHAR i = len, *r = usbMsgPtr;
        if(usbMsgFlags & USB_FLG_MSGPTR_IS_ROM){    /* ROM data */
            while(i--){
                UCHAR c = PRG_RDB(r);    /* assign to UCHAR size variable to enforce byte ops */
                *data++ = c;
                r++;
            }
        }else{                  /* RAM data */
            while(i--)
                *data++ = *r++;
        }
        usbMsgPtr = r;
        return len;
#if USB_CFG_IMPLEMENT_FN_READ
    }else{
        if(len != 0)    /* don't bother app with 0 sized reads */
            return usbFunctionRead(data, len);
        return 0;
    }
#endif
}

void usbProcessRx(UCHAR *data, UCHAR len)
{
	usbRequest_t *rq = (void *)data;
	UCHAR replyLen = 0;
    UCHAR flags = USB_FLG_USE_DEFAULT_RW;
/* We use if() cascades because the compare is done byte-wise while switch()
 * is UINT-based. The if() cascades are therefore more efficient.
 */

#if USB_CFG_IMPLEMENT_FN_WRITEOUT
    if(usbRxToken & 0x80)
    {
        usbFunctionWriteOut(data, len);
        return;
    }
    if(usbRxToken == (UCHAR)(USBPID_SETUP & 0x7f))	/* MSb contains endpoint (== 0) */
    {
#else
    if(usbRxToken == (UCHAR)USBPID_SETUP)
    {
#endif
        if(len == 8)
        {   /* Setup size must be always 8 bytes. Ignore otherwise. */
            UCHAR type = rq->bmRequestType & USBRQ_TYPE_MASK;
            if(type == USBRQ_TYPE_STANDARD)
            {
                UCHAR *replyData = usbTxBuf + 9; /* there is 3 bytes free space at the end of the buffer */
                replyData[0] = 0;   /* common to USBRQ_GET_STATUS and USBRQ_GET_INTERFACE */

                switch(rq->bRequest)
                {
                	case USBRQ_GET_STATUS:	/* 0 */
#if USB_CFG_IS_SELF_POWERED
                  		UCHAR __attribute__((__unused__)) recipient = rq->bmRequestType & USBRQ_RCPT_MASK;  /* assign arith ops to variables to enforce byte size */
                    	if(recipient == USBRQ_RCPT_DEVICE)
                        	replyData[0] =  USB_CFG_IS_SELF_POWERED;
#endif
#if USB_CFG_HAVE_INTRIN_ENDPOINT && USB_CFG_IMPLEMENT_HALT
                    	if(usbHalted1 && recipient == USBRQ_RCPT_ENDPOINT && rq->wIndex.bytes[0] == 0x81)   /* request status for endpoint 1 */
                        	replyData[0] = 1;
#endif
                    	replyData[1] = 0;
                    	replyLen = 2;
                        break;

                	case USBRQ_CLEAR_FEATURE:	/* 1 */
                      	break;

                    case USBRQ_SET_FEATURE:	/* 3 */
                      	break;

                	case USBRQ_SET_ADDRESS:    /* 5 */
                    	usbNewDeviceAddr = rq->wValue.bytes[0];
                        break;

                	case USBRQ_GET_DESCRIPTOR: /* 6 */
                    	flags = USB_FLG_MSGPTR_IS_ROM | USB_FLG_USE_DEFAULT_RW;
                    	switch(rq->wValue.bytes[1])
                        {
                        	case 1:	/* descriptor type requested */
                        		replyLen = sizeof(USBDeviceDescriptor);
                        		replyData = (UCHAR *)&USBDeviceDescriptor;
                                break;

                        	case 2:
                        		replyLen = sizeof(USB_CDC_CommPort_Descriptors_Packet_t);
                        		replyData = (UCHAR *)&USBCDCCommPortDescriptorsPacket;
                                break;

                        	case 3: /* string descriptor */
                              	switch(rq->wValue.bytes[0])
                                {
                                	case 0:   /* descriptor index */
                            			replyLen = usbDescrString0[0];
                            			replyData = (UCHAR *)usbDescrString0;
                                        break;
#if USB_CFG_VENDOR_NAME_LEN
                                	case 1:
                            			replyLen = usbDescrString1[0];
                            			replyData = (UCHAR *)usbDescrString1;
                                        break;
#endif
#if USB_CFG_DEVICE_NAME_LEN
                                	case 2:
                            			replyLen = usbDescrString2[0];
                            			replyData = (UCHAR *)usbDescrString2;
                                        break;
#endif
#if USB_CFG_SERIAL_NUMBER_LENGTH
                                	case 3:
                            			replyLen = 2 * USB_CFG_SERIAL_NUMBER_LENGTH + 2;
                            			replyData = (UCHAR *)usbCfgSerialNumberStringDescriptor;
                                        break;
#endif
                        		}
                    			break;

#if USB_CFG_HID_REPORT_DESCRIPTOR_LENGTH
                        	case USBDESCR_HID:	/* 0x21 */
                        		replyLen = 9;
                        		replyData = (UCHAR *)usbDescrConfig + 18;
                                break;

                        	case USBDESCR_HID_REPORT:   /* 0x22 */
                        		replyLen = USB_CFG_HID_REPORT_DESCRIPTOR_LENGTH;
                        		replyData = (UCHAR *)usbHidReportDescriptor;
                                break;                    	
#endif
                        }
                    	break;

                	case USBRQ_SET_DESCRIPTOR:		/* 7 */
                      	break;

                	case USBRQ_GET_CONFIGURATION:	/* 8 */
                    	replyLen = 1;
                    	replyData = &usbConfiguration;  /* send current configuration value */
                        break;

                	case USBRQ_SET_CONFIGURATION:	/* 9 */
                    	usbConfiguration = rq->wValue.bytes[0];
#if USB_CFG_IMPLEMENT_HALT
                    	usbHalted1 = 0;
#endif
                		break;
                	
                	case USBRQ_GET_INTERFACE:      /* 10 */
                    	replyLen = 1;
                        break;

#if USB_CFG_HAVE_INTRIN_ENDPOINT
#if USB_CFG_IMPLEMENT_HALT
                	case USBRQ_CLEAR_FEATURE:
                	case USBRQ_SET_FEATURE:   /* 1|3 */
                    	if(rq->wValue.bytes[0] == 0 && rq->wIndex.bytes[0] == 0x81)	/* feature 0 == HALT for endpoint == 1 */
                        {
                        	usbHalted1 = rq->bRequest - 1;
                        	if(usbHalted1)
                            {
                            	usbTxBuf1[0] = USBPID_STALL;
                            	usbTxLen1 = 2;      /* length including sync byte */
                        	}
                        	usbTxPacketCnt1 = 0;    /* reset data toggling for interrupt endpoint */
                    	}
                        break;
#endif
                	case USBRQ_SET_INTERFACE:      /* 11 */
                    	usbTxPacketCnt1 = 0;        /* reset data toggling for interrupt endpoint */
#if USB_CFG_IMPLEMENT_HALT
                    	usbHalted1 = 0;
#endif
#endif
                        break;

                	case USBRQ_SYNCH_FRAME: /* 12 */
                      	break;
                }
                usbMsgPtr = replyData;
                if(!rq->wLength.bytes[1] && replyLen > rq->wLength.bytes[0])  /* max length is in */
                    replyLen = rq->wLength.bytes[0];
            }
            else	/* not a standard request -- must be vendor or class request */
            {
                replyLen = usbFunctionSetup(data);
#if USB_CFG_IMPLEMENT_FN_READ || USB_CFG_IMPLEMENT_FN_WRITE
                if(replyLen == 0xff)	/* use user-supplied read/write function */
                {
                    if((rq->bmRequestType & USBRQ_DIR_MASK) == USBRQ_DIR_DEVICE_TO_HOST)
                        replyLen = rq->wLength.bytes[0];    /* IN transfers only */
                    flags = 0;  /* we have no valid msg, use user supplied read/write functions */
                }
#endif
            }
        }
        /* make sure that data packets which are sent as ACK to an OUT transfer are always zero sized */
    }
    else	/* DATA packet from out request */
    {
#if USB_CFG_IMPLEMENT_FN_WRITE
        if(!(usbMsgFlags & USB_FLG_USE_DEFAULT_RW))
        {
            UCHAR rval = usbFunctionWrite(data, len);
            replyLen = 0xff;
            if(rval == 0xff)	/* an error occurred */
            {
                /* usbMsgLen = 0xff; cancel potentially pending ACK [has been done by ASM module when OUT token arrived] */
                usbTxBuf[0] = USBPID_STALL;
                usbTxLen = 2;       /* length including sync byte */
            }
            else
              	if(rval != 0)    /* This was the final package */
                	replyLen = 0;       /* answer with a zero-sized data packet */
            flags = 0;    /* start with a DATA1 package, stay with user supplied write() function */
        }
#else
        replyLen = 0;      /* send zero-sized block as ACK */
#endif
    }

    usbMsgFlags = flags;
    usbMsgLen = replyLen;
}

/* ------------------------------------------------------------------------- */

static void usbBuildTxBlock(void)
{
UCHAR       wantLen, len, txLen, token;

    wantLen = usbMsgLen;
    if(wantLen > 8)
        wantLen = 8;
    usbMsgLen -= wantLen;
    token = USBPID_DATA1;
    if(usbMsgFlags & USB_FLG_TX_PACKET)
        token = USBPID_DATA0;
    usbMsgFlags++;
    len = usbRead(usbTxBuf + 1, wantLen);
    if(len <= 8){           /* valid data packet */
        usbCrc16Append(&usbTxBuf[1], len);
        txLen = len + 4;    /* length including sync byte */
        if(len < 8)         /* a partial package identifies end of message */
            usbMsgLen = 0xff;
    }else{
        token = USBPID_STALL;
        txLen = 2;          /* length including sync byte */
        usbMsgLen = 0xff;
    }
    usbTxBuf[0] = token;
    usbTxLen = txLen;
}

static inline UCHAR isNotSE0(void)
{
UCHAR   rval;
/* We want to do
 *     return (USBIN & USBMASK);
 * here, but the compiler does UINT-expansion acrobatics.
 * We can avoid this by assigning to a UCHAR-sized variable.
 */
    rval = USBIN & USBMASK;
    return rval;
}

/* ------------------------------------------------------------------------- */

void    usbPoll(void)
{
UCHAR   len;

    if((len = usbRxLen) > 0){
/* We could check CRC16 here -- but ACK has already been sent anyway. If you
 * need data integrity checks with this driver, check the CRC in your app
 * code and report errors back to the host. Since the ACK was already sent,
 * retries must be handled on application level.
 * unsigned crc = usbCrc16((UCHAR *)(unsigned)(usbAppBuf + 1), usbRxLen - 3);
 */
        len -= 3;       /* remove PID and CRC */
        if(len < 128){  /* no overflow */
            converter_t appBuf;
            appBuf.ptr = (UCHAR *)usbRxBuf;
            appBuf.bytes[0] = usbAppBuf;
            appBuf.bytes[0]++;
            usbProcessRx(appBuf.ptr, len);
        }
#if USB_CFG_HAVE_FLOWCONTROL
        if(usbRxLen > 0)    /* only mark as available if not inactivated */
            usbRxLen = 0;
#else
        usbRxLen = 0;   /* mark rx buffer as available */
#endif
    }
    if(usbMsgLen != 0xff){  /* transmit data pending? */
        if(usbTxLen < 0)    /* transmit system idle */
            usbBuildTxBlock();
    }
    if(isNotSE0()){ /* SE0 state */
        usbIsReset = 0;
    }else{
        /* check whether SE0 lasts for more than 2.5us (3.75 bit times) */
        if(!usbIsReset){
            UCHAR i;
            for(i=100;i;i--){
                if(isNotSE0())
                    goto notUsbReset;
            }
            usbIsReset = 1;
            usbNewDeviceAddr = 0;
            usbDeviceAddr = 0;
#if USB_CFG_IMPLEMENT_HALT
            usbHalted1 = 0;
#endif

notUsbReset:;
        }
    }
}

/* ------------------------------------------------------------------------- */

void    usbInit(void)
{	
    usbInputBuf = (UCHAR)usbRxBuf[0];
    usbAppBuf = (UCHAR)usbRxBuf[1];
#if USB_INTR_CFG_SET != 0
    USB_INTR_CFG |= USB_INTR_CFG_SET;
#endif
#if USB_INTR_CFG_CLR != 0
    USB_INTR_CFG &= ~(USB_INTR_CFG_CLR);
#endif
    USB_INTR_ENABLE |= (1 << USB_INTR_ENABLE_BIT);
}

/* ------------------------------------------------------------------------- */
